drop table category;
\o tutorial.txt
\qecho ''
\qecho ''
\qecho 'SAMPLE TREE DATA STRUCTURE FOR THE TUTORIAL:'
\qecho
\qecho '            category 1'
\qecho '           /          \\'
\qecho '          /            \\'
\qecho '   category 1.2     category 1.2'
\qecho '        |'
\qecho '        |'
\qecho '   category 1.2.1'
\qecho ''
\qecho ''
\qecho 'this is the final structure that i wish to have. '
\qecho 'the contents of the tree nodes are in reality records of a separate table.'
\qecho 'By the help of the library i will connect these records to a tree '
\qecho 'structure that i will create and control by the library'
\qecho ''
\qecho ''
\qecho 'there is more than one way to create tree hierarchies '
\qecho 'for this tutorial i will chose one of them '
\qecho 'in brief: first of all i will create nodes '
\qecho 'then i will set the hierarchical relationships'
\qecho 'between the nodes'
\qecho ''
\qecho 'CREATE SAMPLE DATA TABLE'
\qecho 'In this table we have only the plain data'
\qecho 'without hierarchical information'
\qecho 'the nodes data for the tutorial are simply varchar fields, in a real situation'
\qecho 'it is possible to be any valid record set' 
\qecho '' 
\qecho '#CREATE TABLE category (id integer primary key, cat_name varchar);'
CREATE TABLE category (id integer primary key, cat_name varchar);
\qecho ''
\qecho '#insert into category values (101, ''category 1'');'
insert into category values (101, 'category 1');
\qecho '#insert into category values (102, ''category 1.1'');'
insert into category values (102, 'category 1.1');
\qecho '#insert into category values (103, ''category 1.2'');'
insert into category values (103, 'category 1.2');
\qecho '#insert into category values (104, ''category 1.2.1'');'
insert into category values (104, 'category 1.2.1');
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho 'CREATE THE TREE STRUCTURE'
\qecho '-------------------------'
\qecho ''
\qecho 'create tree'
\qecho 'SELECT tree.create_tree(\'t1\', \'categories hierarchy 1\');'
SELECT tree.create_tree('t1', 'categories hierarchy 1');
\qecho ''
\qecho 'add nodes without structure'
\qecho ''
\qecho 'for the manipulation of nodes we need a node identificator.'  
\qecho 'By terms of library we have two ids'
\qecho 'a global node id that is unique for all trees registered in library'
\qecho 'and an id that is unique inside the tree, named as node_sn'
\qecho 'in future versions of the library the use of global id will be reduced '
\qecho 'i would propose the use of node_sn when that is possible'
\qecho 'the global ids are automatically assigned by the library'
\qecho 'there is no way to have control over this ids'
\qecho 'for the node_sn we have options similar to the postgres table serial ids '
\qecho 'that are connected to sequences'
\qecho 'it is possible to take a free node_sn with the use of:'
\qecho 'tree.get_node_next_sn(<TREE_NAME>)'
\qecho 'alternately you can use your own numbering schema and continue by setting '
\qecho 'the next free node_sn with:'
\qecho 'SELECT tree.set_node_next_sn(<TREE_NAME>,<VALUE>)'
\qecho 'if you wand to see the status of node_sn sequence, you will make the use of:'
\qecho 'SELECT * from tree.get_node_sn_sequence(<TREE_NAME>)'
\qecho ''
\qecho 'for this example i will use my own numbering schema with the use of: '
\qecho 'tree.add_orphan_sn_node(<tree_name>,<node_sn>,<data_id>) '
\qecho 'The arguments of tree.add_orphan_sn_node are: '
\qecho '1) tree_name: tree name '
\qecho '2) node_sn  : serial number of node (unique per tree)'
\qecho '3) data_id  : the node data id (the id of row at category table)'
\qecho ''
\qecho ''
\qecho '#SELECT tree.add_orphan_sn_node(''t1'',1,101);'
SELECT tree.add_orphan_sn_node('t1',1,101);
\qecho '#SELECT tree.add_orphan_sn_node(''t1'',2,102);'
SELECT tree.add_orphan_sn_node('t1',2,102);
\qecho '#SELECT tree.add_orphan_sn_node(''t1'',3,103);'
SELECT tree.add_orphan_sn_node('t1',3,103);
\qecho '#SELECT tree.add_orphan_sn_node(''t1'',4,104);'
SELECT tree.add_orphan_sn_node('t1',4,104);
\qecho ''
\qecho '#SELECT tree.set_node_next_sn(''t1'',10)'
SELECT tree.set_node_next_sn('t1',10);
\qecho ''
\qecho ''
\qecho ''
\qecho 'add hierarchical information with the use of: '
\qecho 'SELECT tree.set_root_node(<tree_name>,<node_sn>)'
\qecho 'and'
\qecho 'SELECT tree.set_parentship('t1'<parent_node_sn>,<child_node_sn>)'
\qecho ''
\qecho ''
\qecho #SELECT tree.set_root_node('t1',1);
SELECT tree.set_root_node('t1',1);
\qecho #SELECT tree.set_parentship('t1'1,2);
SELECT tree.set_parentship('t1',1,2);
\qecho #tree.set_parentship('t1',1,3);
SELECT tree.set_parentship('t1',1,3);
\qecho #SELECT tree.set_parentship('t1',2,4); 
SELECT tree.set_parentship('t1',2,4);
\qecho ''
\qecho 'the tree structure is registered within the library'
\qecho 'now it is possible to access the tree by the help of '
\qecho 'the tree node retrival library functions '
\qecho ''
\qecho ''
\qecho 'A graphical representation of the tree structure that is finally created:'
\qecho ''
\qecho '           (1->101)'
\qecho '           /     \\'
\qecho '        (2->102)  (3->103)'
\qecho '          |'
\qecho '        (4->104)'
\qecho ' '
\qecho ' (a,b) : (node_id, category.id)'
\qecho ''
\qecho ''
\qecho 'for information about the data structures that the functions return'
\qecho 'look at functions.txt'
\qecho ''        
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho ''
\qecho 'NODE RETRIVAL'
\qecho '-------------'
\qecho ''
\qecho 'get the tree nodes'
\qecho
\qecho SELECT * FROM tree.get_nodes('t1');
\qecho SELECT * FROM tree.get_leaf_nodes('t1');
\qecho SELECT * FROM tree.get_internal_nodes('t1');
\qecho ''
\qecho 'get the childs of a node'
\qecho 'SELECT * from tree.get_childs(1);'
\qecho
\qecho 'get the parent node'
\qecho 'SELECT tree.get_parent(2);'
\qecho 'get the path from node to root'
\qecho 'SELECT * FROM tree.get_path(4,true);'
\qecho
\qecho 'combine the real data with the tree  structure.'
\qecho 'SELECT * FROM public.category where id in   (SELECT node_data_id FROM tree.get_leaf_nodes(\'t1\'));'
\qecho 'SELECT * FROM public.category WHERE id = (tree.get_parent(2)).id;'
\qecho ''
\qecho ''
\qecho 'TREE TRAVERSAL'
\qecho '--------------'
\qecho ''
\qecho 'simple traversal'
\qecho 'SELECT * from tree.traverse_tree(\'t1\')';
\qecho
\qecho 'traversal with level'
\qecho 'SELECT * from tree.traverse_tree_level(\'t1\');'
\qecho
\qecho 'traversal with groups'
\qecho 'SELECT * from tree.traverse_group(1,true);';
\qecho
\qecho '#SELECT' 
\qecho ' CASE'
\qecho '  WHEN group_tag = 1 THEN \'<ul>\''
\qecho '  WHEN group_tag = 2 THEN \'</ul>\''
\qecho '  WHEN group_tag is null THEN  \'<li>\'  || node_data_id  || \'</li>\''
\qecho ' END AS tags'
\qecho ' from tree.traverse_group(1,8,true) t ;'

SELECT 
 CASE
  WHEN group_tag = 1 THEN '<ul>'
  WHEN group_tag = 2 THEN '</ul>'
  WHEN group_tag is null THEN  '<li>'  || node_data_id  || '</li>'
 END AS tags
 from tree.traverse_group(1,8,true) t ;

\qecho ''
\qecho ''
\qecho '#SELECT' 
\qecho ' CASE'
\qecho '  WHEN group_tag = 1 THEN \'<ul>\''
\qecho '  WHEN group_tag = 2 THEN \'</ul>\''
\qecho '  WHEN group_tag is null THEN  \'<li>\'  || d.cat_name  || \'</li>\''
\qecho ' END AS tags'
\qecho ' from tree.traverse_group(1,8,true) t LEFT JOIN public.category d ON (t.node_data_id =  d.id) order by traversal_index ;'
\qecho ''
SELECT
 CASE
 WHEN group_tag = 1 THEN '<ul>'
 WHEN group_tag = 2 THEN '</ul>'
 WHEN group_tag is null THEN  '<li>'  || d.cat_name  || '</li>'
 END AS tags
 from tree.traverse_group(1,8,true) t LEFT JOIN public.category d ON (t.node_data_id =  d.id) order by traversal_index ;
\qecho '' 
\qecho 'HTML RENDERING:'
\qecho ''
\qecho '       *  category 1'
\qecho '             o category 2'
\qecho '                + category 4'
\qecho '             o category 3'
\qecho ''
\qecho  ''
\qecho  ''
\qecho  ''
\qecho  ''
\qecho 'SELECT '
\qecho ' CASE'
\qecho '  WHEN group_tag = 1 THEN \'<ul>\' '
\qecho '  WHEN group_tag = 2 THEN \'</ul>\' '
\qecho ' END AS grou_tags, d.cat_name  '
\qecho ' from tree.traverse_group(1,8,true) t LEFT JOIN public.category d ON (t.node_data_id =  d.id) order by traversal_index ;'
\qecho ''
SELECT 
 CASE
  WHEN group_tag = 1 THEN '<ul>'
  WHEN group_tag = 2 THEN '</ul>'
 END AS grou_tags, d.cat_name 
 from tree.traverse_group(1,8,true) t LEFT JOIN public.category d ON (t.node_data_id =  d.id) order by traversal_index ;
\qecho ''
\qecho ''
\qecho '' 
\qecho ''
\qecho 'for information about the data structures that the functions return'
\qecho 'look at functions.txt'
\qecho '' 
\qecho '' 
\qecho '' 
\qecho '' 
\qecho ''
\qecho 'METADATA '
\qecho '--------'
\qecho ''
\qecho 'SELECT tree.set_node_name(3,\'my_interesting_node\');'
\qecho 'SELECT tree.get_node(\'my_interesting_node\');'
\qecho 'SELECT tree.get_node_id(\'my_interesting_node\');'
\qecho ''
\qecho ''
\qecho ''
\qecho 'VIEWS SAMPLE '
\qecho '------------'
\qecho ''
\qecho ''
\qecho 'CREATE VIEW public.category_view AS' 
\qecho 'SELECT td.*, '
\qecho ' n.id as treenode_id,'
\qecho ' n.weight as treenode_weight,'
\qecho ' n.internal_weight as treenode_internal_weight,'
\qecho ' n.level as treenode_level'
\qecho 'from public.category td'
\qecho 'JOIN tree.get_nodes(\'t1\') n ON (n.node_data_id = td.id);'
