






 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION tree_util.create_user(arg_user varchar) returns integer AS $$
 DECLARE
 rec record;
 p_user_id integer;
 BEGIN
    
    select into rec * FROM tree_util.access_user WHERE username = arg_user;
 	if found then
 		p_user_id := rec.id;
 	else 
 		p_user_id := nextval('tree_util.access_user_id_seq');
 		insert into tree_util.access_user values (p_user_id,arg_user);
 	end if;
 	return p_user_id;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 

CREATE OR REPLACE FUNCTION tree_util.create_role(arg_role varchar) returns integer AS $$
 DECLARE
 rec record;
 p_role_id integer;
 BEGIN
    
    select into rec * FROM tree_util.access_role WHERE rolename = arg_role;
 	if found then
 		p_role_id := rec.id;
 	else 
 		p_role_id := nextval('tree_util.access_role_id_seq');
 		insert into tree_util.access_role values (p_role_id,arg_role);
 	end if;
 	return p_role_id;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 

CREATE OR REPLACE FUNCTION  tree.add_user_to_role(arg_tree varchar, arg_user varchar, arg_role varchar) RETURNS boolean AS $$
 DECLARE
 rec record;
 p_tree_id integer;
 p_user_id integer;
 p_role_id integer;
 BEGIN
 
 	p_tree_id := tree_util.get_tree_id(arg_tree);
 	
 	if p_tree_id is null then
 		return false;
 	end if;
 	
 	p_user_id := tree_util.create_user(arg_user);
 	p_role_id := tree_util.create_role(arg_role);
 	
 	
 	
 	SELECT into rec * FROM tree_util.user_to_role WHERE tree_id = p_tree_id and user_id = p_user_id and role_id = p_role_id;
 	if found then
 	return true;
 	end if;
 	
 	insert into tree_util.user_to_role values (default,p_tree_id, p_user_id, p_role_id);
 	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 
 


 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.add_node_to_role( arg_node bigint, arg_role varchar) RETURNS boolean AS $$
 DECLARE
 p_role_id integer;
 rec record;
 BEGIN
 
 	p_role_id := tree_util.create_role(arg_role);
 	
 	SELECT into rec * FROM tree_util.node_to_role WHERE  node_id = arg_node and role_id = p_role_id;
 	if found then
 	return true;
 	end if;
 	
 	SELECT into rec * FROM tree_util.tree_node where id = arg_node;
 	if not found then
 	return false;
 	end if;
 	
 	insert into tree_util.node_to_role values (default, arg_node,p_role_id);
 	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.add_node_to_role(arg_tree varchar, arg_node varchar, arg_role varchar) RETURNS boolean AS $$
 DECLARE
 p_role_id integer;
 p_tree_id integer;
 rec1 record;
 rec2 record;
 BEGIN
 	
 	
 	p_tree_id := tree_util.get_tree_id(arg_tree);
 	
 	if p_tree_id is null then
 		return false;
 	end if;
 	
 	p_role_id := tree_util.create_role(arg_role);
 	
 	
 	SELECT into rec1 * FROM tree_util.tree_node where tree_id = p_tree_id and node_name = arg_node;
 	if not found then
 	return false;
 	end if;
 	
 	
 	SELECT into rec2 * FROM tree_util.node_to_role WHERE  node_id = rec1.id and role_id = p_role_id;
 	if found then
 	return true;
 	end if;
 	
 	insert into tree_util.node_to_role values (default, rec1.id,p_role_id);
 	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.add_node_to_role(arg_tree varchar, arg_node_sn bigint, arg_role varchar) RETURNS boolean AS $$
 DECLARE
 p_role_id integer;
 p_tree_id integer;
 rec1 record;
 rec2 record;
 BEGIN
 	
 	
 	p_tree_id := tree_util.get_tree_id(arg_tree);
 	
 	if p_tree_id is null then
 		return false;
 	end if;
 	
 	p_role_id := tree_util.create_role(arg_role);
 	
 	
 	SELECT into rec1 * FROM tree_util.tree_node where tree_id = p_tree_id and node_sn = arg_node_sn;
 	if not found then
 	return false;
 	end if;
 	
 	
 	SELECT into rec2 * FROM tree_util.node_to_role WHERE  node_id = rec1.id and role_id = p_role_id;
 	if found then
 	return true;
 	end if;
 	
 	insert into tree_util.node_to_role values (default, rec1.id,p_role_id);
 	return true;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

 CREATE OR REPLACE FUNCTION  tree.drop_role_associations(arg_role varchar) RETURNS boolean AS $$
 DECLARE
 BEGIN
 	DELETE FROM  tree_util.access_role WHERE rolename = arg_role;
 	if found then
 		return true;
 	end if;
 	return false;
 END
 $$ VOLATILE LANGUAGE plpgsql;
 
 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.drop_role_user_associations(arg_role varchar) RETURNS boolean AS $$
 DECLARE
 	p_role_id integer;
 	rec record;
 BEGIN
 	SELECT into rec id from tree_util.access_role;
 	if not found then
 		return false;
 	end if;
 	p_role_id :=rec.id;
 	
 	DELETE FROM  tree_util.user_to_role WHERE role_id = p_role_id;
 	if found then
 		return true;
 	end if;
 	return false;
 END
 $$ VOLATILE LANGUAGE plpgsql;

 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------

 CREATE OR REPLACE FUNCTION  tree.drop_role_node_associations(arg_role varchar) RETURNS boolean AS $$
 DECLARE
 	p_role_id integer;
 	rec record;
 BEGIN
 	SELECT into rec id from tree_util.access_role;
 	if not found then
 		return false;
 	end if;
 	p_role_id :=rec.id;
 	
 	DELETE FROM  tree_util.node_to_role WHERE role_id = p_role_id;
 	if found then
 		return true;
 	end if;
 	return false;
 END
 $$ VOLATILE LANGUAGE plpgsql;

 
 



 ---------------------------------------------------------------------------
 --
 ---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  tree.is_user_in_role(arg_tree varchar, arg_user varchar, arg_role varchar) RETURNS boolean AS $$
 DECLARE
 rec record;
 BEGIN
 	SELECT into rec 1 FROM tree_util.user_to_role ur
    JOIN tree_util.access_role r ON (r.id = ur.role_id)
    JOIN tree_util.access_user u ON (u.id = ur.user_id)	
 	WHERE  ur.tree_id = arg_tree and u.username = arg_user and r.rolename = arg_role;
 	if found then
 	return true;
 	end if;
 	
 	return false;
 	
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 
 

 

 
 
 CREATE OR REPLACE FUNCTION  tree.get_nodes_with_roles(arg_tree varchar) RETURNS setof tree.tree_node_with_role AS $$
 	SELECT (n.id, n.node_name,n.node_data_id, n.weight, n.internal_weight, n.level, n.parent_id , ar.rolename)::tree.tree_node_with_role
	FROM tree_util.tree_node n 
	JOIN tree_util.tree t ON (n.tree_id = t.id)
	JOIN tree_util.node_to_role nr ON (n.id = nr.node_id)
	JOIN tree_util.access_role ar ON (ar.id = nr.role_id)
	WHERE t.tree_name = $1
 $$ STABLE LANGUAGE sql;
 
 		

 
 
 CREATE OR REPLACE FUNCTION  tree.get_users_with_roles(arg_tree varchar) RETURNS setof tree.user_with_role AS $$
 	SELECT (u.username, ar.rolename)::tree.user_with_role
	FROM tree_util.access_user u 
	JOIN tree_util.user_to_role ur ON (u.id = ur.user_id)
	JOIN tree_util.tree t ON (ur.tree_id = t.id)
	JOIN tree_util.access_role ar ON (ar.id = ur.role_id)
	WHERE t.tree_name = $1
 $$ STABLE LANGUAGE sql;
 

 
 		
 

 
 
 
 




---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION rtree_util.get_traversal_node(arg_node rtree_util.tree_node, arg_traversal_index bigint) RETURNS  rtree.traversal_node AS $$
DECLARE
	tn rtree.traversal_node;
begin
    	tn.id := arg_node.id;
    	tn.node_sn := arg_node.node_sn;
    	tn.node_name := arg_node.node_name;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		tn.parent_sn := arg_node.parent_sn;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION rtree_util.get_traversal_node(arg_node rtree.tree_node, arg_traversal_index bigint) RETURNS  rtree.traversal_node AS $$
DECLARE
	tn rtree.traversal_node;
begin
    	tn.id := arg_node.id;
    	tn.node_sn := arg_node.node_sn;
    	tn.node_name := arg_node.node_name;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		tn.parent_sn := arg_node.parent_sn;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION rtree_util.get_traversal_group(arg_node rtree.tree_node, arg_traversal_index integer, arg_group_tag integer, arg_node_traversal_index integer) RETURNS  rtree.traversal_group AS $$
DECLARE
	tn rtree.traversal_group;
begin
		tn.group_tag := arg_group_tag;
		tn.node_sn := arg_node.node_sn;
    	tn.node_id := arg_node.id;
    	tn.node_name := arg_node.node_name;
    	tn.node_traversal_index := arg_node_traversal_index;
    	tn.traversal_index := arg_traversal_index;
    	tn.node_data_id := arg_node.node_data_id;
        tn.weight := arg_node.weight;
    	tn.internal_weight := arg_node.internal_weight;
    	tn.level := arg_node.level;
		tn.parent_id := arg_node.parent_id;
		tn.parent_sn := arg_node.parent_sn;
		return  tn;
END
$$ IMMUTABLE LANGUAGE plpgsql;



---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
CREATE OR REPLACE FUNCTION rtree_util.get_tree_id(arg_tree_name varchar) RETURNS  integer AS $$
	SELECT id from rtree_util.tree WHERE tree_name = $1;
$$ STABLE LANGUAGE sql;

CREATE OR REPLACE FUNCTION rtree_util.get_tree_name(arg_tree_id bigint) RETURNS  varchar AS $$
	SELECT tree_name from rtree_util.tree WHERE id = $1;
$$ STABLE LANGUAGE sql;
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 


 CREATE OR REPLACE FUNCTION  rtree_util.get_root(arg_tree_id integer) RETURNS  rtree.tree_node AS $$
   SELECT (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn where level = 0 and tree_id = $1;
 $$ STABLE LANGUAGE sql;
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 CREATE OR REPLACE FUNCTION  rtree_util.get_root_id(arg_tree_id integer) RETURNS  bigint AS $$
   SELECT  id FROM rtree_util.tree_node where level = 0 and tree_id = $1;
 $$ STABLE LANGUAGE sql;




---------------------------------------------------------------------------
--
---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION rtree_util.get_tree_node_id(arg_tree_name varchar, arg_node_sn bigint) RETURNS bigint AS $$
	SELECT tn.id from rtree_util.tree_node tn JOIN rtree_util.tree t ON  (tn.node_sn = $2 and tn.tree_id = t.id and t.tree_name = $1);
$$ STABLE LANGUAGE  sql;

CREATE OR REPLACE FUNCTION rtree_util.get_tree_node(arg_tree_name varchar, arg_node_sn bigint) RETURNS rtree_util.tree_node AS $$
	SELECT tn from rtree_util.tree_node tn JOIN rtree_util.tree t ON  (tn.node_sn = $2 and tn.tree_id = t.id and t.tree_name = $1);
$$ STABLE LANGUAGE  sql;


CREATE OR REPLACE FUNCTION rtree_util.get_tree_node_id(arg_tree_id bigint, arg_node_sn bigint) RETURNS bigint AS $$
	SELECT tn.id from rtree_util.tree_node tn WHERE tn.node_sn = $2 and tn.tree_id = $1;
$$ STABLE LANGUAGE  sql;

CREATE OR REPLACE FUNCTION rtree_util.get_tree_node(arg_tree_id bigint, arg_node_sn bigint) RETURNS rtree_util.tree_node AS $$
	SELECT tn from rtree_util.tree_node tn WHERE  tn.node_sn = $2 and tn.tree_id = $1;
$$ STABLE LANGUAGE  sql;



---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
-- tree
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
---------------------------------------------------------------------------
--------------------------------------------------------------------------- 





---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
CREATE OR REPLACE FUNCTION  rtree.get_childs(node_id bigint) RETURNS  setof  rtree.tree_node AS $$
  	SELECT (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn
    WHERE parent_id = $1
	ORDER BY tn.weight, tn.internal_weight
 $$ STABLE LANGUAGE sql;
 
 
 CREATE OR REPLACE FUNCTION  rtree_util.get_childs(node_id bigint) RETURNS  setof  rtree_util.tree_node AS $$
  	SELECT (tn.*) FROM rtree_util.tree_node tn
    WHERE parent_id = $1
	ORDER BY tn.weight, tn.internal_weight
 $$ STABLE LANGUAGE sql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
CREATE OR REPLACE FUNCTION  rtree.get_nodes_for_data(tree_name varchar, node_data_id bigint) RETURNS  setof  rtree.tree_node AS $$
  	SELECT (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn
  	WHERE 
  	tree_id = rtree_util.get_tree_id($1) AND node_data_id = $2
	ORDER BY tn.weight, tn.internal_weight
 $$ STABLE LANGUAGE sql; 
 
 
 
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree.get_leaf_nodes(tree_name varchar) RETURNS  setof  rtree.tree_node AS $$
    SELECT (n.*)::rtree.tree_node from rtree_util.tree_node n  
    JOIN (
    (SELECT distinct id as family_id from rtree_util.tree_node tn WHERE tn.tree_id = rtree_util.get_tree_id($1))
    except
    (SELECT distinct parent_id  as family_id from  rtree_util.tree_node)) f 
    ON ( n.id = f.family_id)
 $$ STABLE LANGUAGE sql;
 


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
  
 
 CREATE OR REPLACE FUNCTION  rtree.get_internal_nodes(tree_name varchar) RETURNS  setof  rtree.tree_node AS $$
    SELECT (n.*)::rtree.tree_node  from rtree_util.tree_node n  
    JOIN (SELECT distinct parent_id as family_id  from  rtree_util.tree_node) f 
    ON (n.tree_id = rtree_util.get_tree_id($1) and n.id = f.family_id)
 $$ STABLE LANGUAGE sql;
 
 
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  rtree.get_root(tree_name varchar) RETURNS  rtree.tree_node AS $$
   SELECT  (n.*)::rtree.tree_node FROM rtree_util.tree_node n 
   where
   n.level = 0 
   and  n.tree_id = rtree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 
 CREATE OR REPLACE FUNCTION  rtree.get_root_id(tree_name varchar) RETURNS  bigint AS $$
 	SELECT (rtree.get_root($1)).id;
 $$
 language sql stable;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
  CREATE OR REPLACE FUNCTION  rtree.get_parent(node_id bigint) RETURNS  rtree.tree_node AS $$
  	SELECT (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn
  	JOIN rtree_util.tree_node fm ON (tn.id = fm.parent_id)
	WHERE fm.id = $1 ;
 $$ STABLE LANGUAGE sql;

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree.get_node(node_id bigint) RETURNS  rtree.tree_node AS $$
  	SELECT  (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn
	WHERE tn.id = $1;
 $$ STABLE LANGUAGE sql;
 
  CREATE OR REPLACE FUNCTION  rtree_util.get_node_extv(node_id bigint) RETURNS  rtree.tree_node AS $$
  	SELECT  (tn.*)::rtree.tree_node FROM rtree_util.tree_node tn
	WHERE tn.id = $1;
 $$ STABLE LANGUAGE sql;
  
 CREATE OR REPLACE FUNCTION  rtree_util.get_node(node_id bigint) RETURNS  rtree_util.tree_node AS $$
  	SELECT  (tn.*) FROM rtree_util.tree_node tn
	WHERE tn.id = $1;
 $$ STABLE LANGUAGE sql;
 
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

 
 CREATE OR REPLACE FUNCTION  rtree.get_nodes(tree_name varchar) RETURNS  setof  rtree.tree_node AS $$
    SELECT (n.*)::rtree.tree_node from rtree_util.tree_node n  
    WHERE
      n.level is not null AND
      n.tree_id = rtree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 
 CREATE OR REPLACE FUNCTION  rtree.get_all_nodes(tree_name varchar) RETURNS  setof  rtree.tree_node AS $$
  SELECT (n.*)::rtree.tree_node from rtree_util.tree_node n  
    WHERE
      n.tree_id = rtree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree.get_orphan_nodes(tree_name varchar) RETURNS  setof  rtree.tree_node AS $$
    SELECT (n.*)::rtree.tree_node from rtree_util.tree_node n
    WHERE
      n.level is null AND
      n.tree_id =  rtree_util.get_tree_id($1);
 $$ STABLE LANGUAGE sql;
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION rtree.get_node(tree_name varchar, node_name varchar)  RETURNS  rtree.tree_node AS $$
 	select (n.*)::rtree.tree_node  from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_name = $2;
 $$ STABLE LANGUAGE sql;
 
  
 CREATE OR REPLACE FUNCTION rtree_util.get_node(tree_name varchar, node_name varchar)  RETURNS  rtree_util.tree_node AS $$
 	select (n.*)  from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_name = $2;
 $$ STABLE LANGUAGE sql;
  
 CREATE OR REPLACE FUNCTION rtree.get_node(tree_name varchar, node_sn bigint)  RETURNS  rtree.tree_node AS $$
 	select (n.*)::rtree.tree_node  from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_sn = $2;
 $$ STABLE LANGUAGE sql;

 
  CREATE OR REPLACE FUNCTION rtree_util.get_node(tree_name varchar, node_sn bigint)  RETURNS  rtree_util.tree_node AS $$
 	select n.*  from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_sn = $2;
 $$ STABLE LANGUAGE sql;
 
 
 
 CREATE OR REPLACE FUNCTION rtree.get_node_id(tree_name varchar, node_name varchar)  RETURNS bigint AS $$
 
 	select n.id from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_name = $2;

 $$ STABLE LANGUAGE sql;
 
  CREATE OR REPLACE FUNCTION rtree.get_node_id(tree_name varchar, node_sn bigint)  RETURNS bigint AS $$
 
 	select n.id from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and 
 	n.node_sn = $2;

 $$ STABLE LANGUAGE sql;
 
 
 

CREATE OR REPLACE FUNCTION rtree.get_node_metadata(node_id bigint)  RETURNS  rtree.tree_node_metadata AS $$
 	select (mt.*)::rtree.tree_node_metadata from rtree_util.tree_node_metadata mt
 	WHERE id = $1;

 $$ STABLE LANGUAGE sql;
 



 
 
 CREATE OR REPLACE FUNCTION rtree.get_named_nodes(tree_name varchar)  RETURNS setof rtree.tree_node AS $$
	select (n.*)::rtree.tree_node  from rtree_util.tree_node n
 	JOIN rtree_util.tree t ON (n.tree_id = t.id) 
 	WHERE t.tree_name = $1 and
 	n.node_name is not null;

 $$ STABLE LANGUAGE sql;
 
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------

 
 
CREATE OR REPLACE FUNCTION  rtree.get_tree(tree_name varchar) RETURNS  rtree.tree AS $$
  	SELECT  (t.*)::rtree.tree FROM rtree_util.tree t
	WHERE t.tree_name = $1 ;
 $$ STABLE LANGUAGE sql;
 
 
 
 
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree.get_trees() RETURNS  setof  rtree.tree AS $$
    SELECT (t.*)::rtree.tree from rtree_util.tree t;
 $$ STABLE LANGUAGE sql;
 
 
 
 
 
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 ---------------------------------------------------------------------------
 
 

---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree.get_path(node_id bigint,  include_node boolean ) RETURNS  setof  rtree.traversal_node AS $$
 DECLARE
 tn rtree_util.tree_node;
 rep rtree.traversal_node;
 nid bigint;
 i integer default 0;
 BEGIN
    nid := node_id;
    
    if include_node = true then
    	i := i + 1;
 		tn := rtree_util.get_node(nid);
 	    rep := rtree_util.get_traversal_node(tn,i);
 		return next rep;
 	end if;
 
 LOOP
 	tn := rtree.get_parent(nid);
 	nid := tn.id;
	if nid is not null then	
		i := i + 1;
		rep := rtree_util.get_traversal_node(tn,i);
		return next rep;
	end if; 	
    EXIT WHEN nid is null;  
 END LOOP;
 
 return;
 
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 CREATE OR REPLACE FUNCTION  rtree.get_path(node_id bigint,  include_node boolean, include_root boolean ) RETURNS  setof  rtree.traversal_node AS $$
 DECLARE
 tn rtree_util.tree_node;
 rep rtree.traversal_node;
 nid bigint;
 i integer default 0;
 BEGIN
    nid := node_id;
    
    if include_node = true then
    	i := i + 1;
 		tn := rtree_util.get_node(nid);
 	    rep := rtree_util.get_traversal_node(tn,i);
 		return next rep;
 	end if;
 
 LOOP
 	tn := rtree.get_parent(nid);
 	nid := tn.id;
 	if not include_root and tn.parent_id is null then
 		nid := null;
 	end if;
	if nid is not null then	
		i := i + 1;
		rep := rtree_util.get_traversal_node(tn,i);
		return next rep;
	end if; 	
    EXIT WHEN nid is null;  
 END LOOP;
 
 return;
 
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 --CREATE OR REPLACE FUNCTION  rtree.get_anchestors(node_id bigint,  node_id bigint ) RETURNS  setof rtree.tree_node AS $$
 --(SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from rtree.get_path($1,false,false) )  
 --INTERSECT 
 --(SELECT id,node_name,node_data_id,weight, internal_weight,level,parent_id from rtree.get_path($2,false,false) );
 --$$ STABLE LANGUAGE sql;
 
 
  CREATE OR REPLACE FUNCTION  rtree.get_anchestor(node_id bigint,  node_id bigint ) RETURNS  setof rtree.tree_node AS $$
 (SELECT id, node_sn, node_name,node_data_id,weight, internal_weight,level,parent_id,parent_sn from rtree.get_path($1,true,false) )  
 INTERSECT 
 (SELECT id,node_sn, node_name,node_data_id,weight, internal_weight,level,parent_id,parent_sn from rtree.get_path($2,true,false) )
 order by level desc limit 1
 ;
 $$ STABLE LANGUAGE sql;
 


 ---------------------------------------------------------------------------
 --rtree_util.traverse_node
 ---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION rtree_util.traverse_node(arg_node rtree.tree_node, arg_max_level integer) RETURNS  setof  rtree.tree_node AS $$
DECLARE
	id1 rtree.tree_node ;
	id2 rtree.tree_node ;
begin
    return next arg_node;
    FOR id1 IN
    SELECT * FROM rtree.get_childs(arg_node.id)
    LOOP
    	IF arg_max_level < 0 OR id1.level <= arg_max_level THEN
         FOR id2 IN
         SELECT * FROM rtree_util.traverse_node(id1,arg_max_level)
         LOOP
            return next id2;
         END LOOP;
        END IF;
    END LOOP;
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
---------------------------------------------------------------------------
-- tree
---------------------------------------------------------------------------
---------------------------------------------------------------------------



---------------------------------------------------------------------------
--
---------------------------------------------------------------------------




CREATE OR REPLACE FUNCTION rtree_util.traverse_node(arg_node_id bigint, arg_max_level integer) RETURNS  setof  rtree.traversal_node AS $$
DECLARE
	rec rtree.tree_node ;
	tn rtree.traversal_node;
	i integer DEFAULT 0;
begin
    FOR rec IN
    SELECT * FROM rtree_util.traverse_node(rtree_util.get_node_extv(arg_node_id), arg_max_level)
    LOOP
    	i := i + 1;
    	tn := rtree_util.get_traversal_node(rec,i);
		return next tn;
    END LOOP;
    return;
END
$$ STABLE LANGUAGE plpgsql;




CREATE OR REPLACE FUNCTION rtree.traverse_node(node_id bigint, max_level integer) RETURNS  setof  rtree.traversal_node AS $$
	SELECT * FROM rtree_util.traverse_node($1,$2);
$$ STABLE LANGUAGE sql;

CREATE OR REPLACE FUNCTION rtree.traverse_node(anode_id bigint) RETURNS  setof  rtree.traversal_node AS $$
	SELECT * FROM rtree_util.traverse_node($1,-1);
$$ STABLE LANGUAGE sql;

 
 CREATE OR REPLACE FUNCTION  rtree.traverse_tree(tree_name varchar) RETURNS  setof  rtree.traversal_node AS $$
 SELECT * FROM rtree_util.traverse_node(rtree.get_root_id($1),-1);
 $$ STABLE LANGUAGE sql;

 
 CREATE OR REPLACE FUNCTION  rtree.traverse_tree(tree_name varchar, max_level integer) RETURNS  setof  rtree.traversal_node AS $$
 SELECT * FROM rtree_util.traverse_node(rtree.get_root_id($1),$2);
 $$ STABLE LANGUAGE sql;
 
 
 
 
 
 
 
 
 
 
 
 
 


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 
 CREATE OR REPLACE FUNCTION  rtree_util.traverse_tree_level(arg_tree_name varchar) RETURNS  setof  rtree.traversal_node AS $$
 DECLARE
 rec RECORD;
 rep rtree.traversal_node;
 i integer DEFAULT 0;
 p_t_id integer; 
 BEGIN
 
   if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return ;
   end if; 
   
   p_t_id := rtree_util.get_tree_id(arg_tree_name);
   if p_t_id is null then
   	RAISE NOTICE 'tree with name % not exists', arg_tree_name;
	return;
   end if;
   
 
   FOR rec IN
      SELECT * FROM
	  (SELECT 
	  n.*, 
      (rtree.get_parent(n.id)) as parent
      FROM rtree_util.tree_node n  
      WHERE
      n.tree_id = p_t_id ) as s
      order by level, (parent).weight, (parent).internal_weight, weight, internal_weight
    LOOP
    	i := i+1;
    	rep := rtree_util.get_traversal_node((rec.id, rec.node_sn, rec.node_name, rec.node_data_id, rec.weight, rec.internal_weight, rec.level, rec.parent_id,rec.parent_sn )::rtree.tree_node,i);
		return next rep;    	
    END LOOP;
    return;
   
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 
  CREATE OR REPLACE FUNCTION  rtree_util.traverse_tree_level(arg_tree_name varchar, arg_max_level integer) RETURNS  setof  rtree.traversal_node AS $$
 DECLARE
 rec RECORD;
 rep rtree.traversal_node;
 i integer DEFAULT 0;
 p_t_id integer; 
 BEGIN
 
  if arg_tree_name is null then
 		RAISE NOTICE 'arg_tree_name IS MANTATORY';
 		return ;
   end if; 
   
   
   p_t_id := rtree_util.get_tree_id(arg_tree_name);
   if p_t_id is null then
   	RAISE NOTICE 'tree with name % not exists', arg_tree_name;
	return;
   end if;
 
   FOR rec IN
      SELECT * FROM
	  (SELECT 
	  n.*, 
      (rtree.get_parent(n.id)) as parent
      FROM rtree_util.tree_node n  
      WHERE
      n.level is not null and
      n.level <= arg_max_level and n.tree_id = p_t_id ) as s
      order by level, (parent).weight, (parent).internal_weight, weight, internal_weight
    LOOP
    	i := i+1;
    	rep := rtree_util.get_traversal_node((rec.id, rec.node_sn, rec.node_name, rec.node_data_id, rec.weight, rec.internal_weight, rec.level, rec.parent_id,rec.parent_sn )::rtree.tree_node,i);
		return next rep;    	
    END LOOP;
    return;
   
 END
 $$ STABLE LANGUAGE plpgsql;
 
 
 
 
  CREATE OR REPLACE FUNCTION  rtree.traverse_tree_level(arg_tree_name varchar, arg_max_level integer) RETURNS  setof  rtree.traversal_node AS $$
	SELECT * FROM rtree_util.traverse_tree_level($1,$2);
  $$ STABLE LANGUAGE sql;
 
CREATE OR REPLACE FUNCTION  rtree.traverse_tree_level(arg_tree_name varchar) RETURNS  setof  rtree.traversal_node AS $$
	SELECT * FROM rtree_util.traverse_tree_level($1);
  $$ STABLE LANGUAGE sql;
  
---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 


CREATE OR REPLACE FUNCTION rtree_util.traverse_group(arg_node rtree.traversal_group, arg_max_level integer) RETURNS  setof  rtree.traversal_group AS $$
DECLARE
	id1 rtree.tree_node ;
	id2 rtree.traversal_group;
	rep rtree.traversal_group;
	rec record;
	p_tag_flag boolean;
begin
    return next arg_node;
  
    SELECT into rec count(*) FROM rtree.get_childs(arg_node.node_id);
    IF rec.count > 0 THEN
      p_tag_flag := true;
      rep := rtree_util.get_traversal_group(null,null,1,null);
      return next rep;
    ELSE 
    	p_tag_flag := false;
    END IF;
    
    FOR id1 IN
    SELECT * FROM rtree.get_childs(arg_node.node_id)
    LOOP
    	IF arg_max_level < 0 OR id1.level <= arg_max_level THEN
         FOR id2 IN
         SELECT * FROM rtree_util.traverse_group( rtree_util.get_traversal_group(id1,null,null,null) ,arg_max_level)
         LOOP
            return next id2;
         END LOOP;
        END IF;
    END LOOP;
    
    IF p_tag_flag THEN
    	rep := rtree_util.get_traversal_group(null,null,2,null);
   		return next rep;
    END IF;
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------
 

CREATE OR REPLACE FUNCTION rtree_util.traverse_group(arg_node_id bigint, arg_max_level integer, get_root_group boolean) RETURNS  setof  rtree.traversal_group AS $$
DECLARE
	rec rtree.traversal_group ;
	i1 integer DEFAULT 0;
	i2 integer DEFAULT 0;
begin
 
 
if get_root_group = true then
 	i1 := i1 + 1;
 	rec := rtree_util.get_traversal_group(null,i1,1,null);
    return next rec;
end if;
 
 FOR rec IN
    SELECT * FROM rtree_util.traverse_group(rtree_util.get_traversal_group(rtree_util.get_node_extv(arg_node_id),null,null,null), arg_max_level)
    LOOP
		if (get_root_group = true or rec.level > 0 or rec.level is null) then
			i1 := i1 + 1;
    		rec.traversal_index := i1;
			if rec.node_id is not null then
    			i2 := i2 + 1;
    			rec.node_traversal_index := i2;
    		else 
    			rec.node_traversal_index := null;
			end if;
		
			return next rec;
		end if;
    END LOOP;
    
    if get_root_group = true then
 		i1 := i1 + 1;
 		rec := rtree_util.get_traversal_group(null,i1,2,null);
   		return next rec;
 	end if;

	return;
	
END
$$ STABLE LANGUAGE plpgsql;


---------------------------------------------------------------------------
--
---------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION rtree.traverse_group(node_id bigint, arg_max_level integer, get_root_group boolean) RETURNS  setof  rtree.traversal_group AS $$
    SELECT * FROM rtree_util.traverse_group($1,$2,$3);
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION rtree.traverse_group(node_id bigint, get_root_group boolean) RETURNS  setof  rtree.traversal_group AS $$
    SELECT * FROM rtree.traverse_group($1,-1,$2)
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION rtree.traverse_group(tree_name varchar, get_root_group boolean) RETURNS  setof  rtree.traversal_group AS $$
    SELECT * FROM rtree.traverse_group(rtree.get_root_id($1),-1,$2)
$$ STABLE LANGUAGE sql;


CREATE OR REPLACE FUNCTION rtree.traverse_group(tree_name varchar, arg_max_level integer, get_root_group boolean) RETURNS  setof  rtree.traversal_group AS $$
    SELECT * FROM rtree.traverse_group(rtree.get_root_id($1),$2,$3)
$$ STABLE LANGUAGE sql;

